@@ -29,6 +29,18 @@
 
 #include "nghttp2_session.h"
 #include "nghttp2_helper.h"
+#include "nghttp2_debug.h"
+
+/* Maximum distance between any two stream's cycle in the same
+   prirority queue.  Imagine stream A's cycle is A, and stream B's
+   cycle is B, and A < B.  The cycle is unsigned 32 bit integer, it
+   may get overflow.  Because of how we calculate the next cycle
+   value, if B - A is less than or equals to
+   NGHTTP2_MAX_CYCLE_DISTANCE, A and B are in the same scale, in other
+   words, B is really greater than or equal to A.  Otherwise, A is a
+   result of overflow, and it is actually A > B if we consider that
+   fact. */
+#define NGHTTP2_MAX_CYCLE_DISTANCE (16384 * 256 + 255)
 
 static int stream_less(const void *lhsx, const void *rhsx) {
   const nghttp2_stream *lhs, *rhs;
@@ -36,8 +48,15 @@ static int stream_less(const void *lhsx, const void *rhsx) {
   lhs = nghttp2_struct_of(lhsx, nghttp2_stream, pq_entry);
   rhs = nghttp2_struct_of(rhsx, nghttp2_stream, pq_entry);
 
-  return lhs->cycle < rhs->cycle ||
-         (lhs->cycle == rhs->cycle && lhs->seq < rhs->seq);
+  if (lhs->cycle == rhs->cycle) {
+    return lhs->seq < rhs->seq;
+  }
+
+  if (lhs->cycle < rhs->cycle) {
+    return rhs->cycle - lhs->cycle <= NGHTTP2_MAX_CYCLE_DISTANCE;
+  }
+
+  return lhs->cycle - rhs->cycle > NGHTTP2_MAX_CYCLE_DISTANCE;
 }
 
 void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
@@ -80,6 +99,7 @@ void nghttp2_stream_init(nghttp2_stream *stream, int32_t stream_id,
   stream->queued = 0;
   stream->descendant_last_cycle = 0;
   stream->cycle = 0;
+  stream->pending_penalty = 0;
   stream->descendant_next_seq = 0;
   stream->seq = 0;
   stream->last_writelen = 0;
@@ -115,9 +135,14 @@ static int stream_subtree_active(nghttp2_stream *stream) {
 /*
  * Returns next cycle for |stream|.
  */
-static uint64_t stream_next_cycle(nghttp2_stream *stream, uint64_t last_cycle) {
-  return last_cycle +
-         stream->last_writelen * NGHTTP2_MAX_WEIGHT / (uint32_t)stream->weight;
+static void stream_next_cycle(nghttp2_stream *stream, uint32_t last_cycle) {
+  uint32_t penalty;
+
+  penalty = (uint32_t)stream->last_writelen * NGHTTP2_MAX_WEIGHT +
+            stream->pending_penalty;
+
+  stream->cycle = last_cycle + penalty / (uint32_t)stream->weight;
+  stream->pending_penalty = penalty % (uint32_t)stream->weight;
 }
 
 static int stream_obq_push(nghttp2_stream *dep_stream, nghttp2_stream *stream) {
@@ -125,15 +150,14 @@ static int stream_obq_push(nghttp2_stream *dep_stream, nghttp2_stream *stream) {
 
   for (; dep_stream && !stream->queued;
        stream = dep_stream, dep_stream = dep_stream->dep_prev) {
-    stream->cycle =
-        stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream_next_cycle(stream, dep_stream->descendant_last_cycle);
     stream->seq = dep_stream->descendant_next_seq++;
 
-    DEBUGF(fprintf(stderr, "stream: stream=%d obq push cycle=%ld\n",
-                   stream->stream_id, stream->cycle));
+    DEBUGF("stream: stream=%d obq push cycle=%d\n", stream->stream_id,
+           stream->cycle);
 
-    DEBUGF(fprintf(stderr, "stream: push stream %d to stream %d\n",
-                   stream->stream_id, dep_stream->stream_id));
+    DEBUGF("stream: push stream %d to stream %d\n", stream->stream_id,
+           dep_stream->stream_id);
 
     rv = nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
     if (rv != 0) {
@@ -160,8 +184,8 @@ static void stream_obq_remove(nghttp2_stream *stream) {
   }
 
   for (; dep_stream; stream = dep_stream, dep_stream = dep_stream->dep_prev) {
-    DEBUGF(fprintf(stderr, "stream: remove stream %d from stream %d\n",
-                   stream->stream_id, dep_stream->stream_id));
+    DEBUGF("stream: remove stream %d from stream %d\n", stream->stream_id,
+           dep_stream->stream_id);
 
     nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
 
@@ -169,6 +193,7 @@ static void stream_obq_remove(nghttp2_stream *stream) {
 
     stream->queued = 0;
     stream->cycle = 0;
+    stream->pending_penalty = 0;
     stream->descendant_last_cycle = 0;
     stream->last_writelen = 0;
 
@@ -190,8 +215,8 @@ static int stream_obq_move(nghttp2_stream *dest, nghttp2_stream *src,
     return 0;
   }
 
-  DEBUGF(fprintf(stderr, "stream: remove stream %d from stream %d (move)\n",
-                 stream->stream_id, src->stream_id));
+  DEBUGF("stream: remove stream %d from stream %d (move)\n", stream->stream_id,
+         src->stream_id);
 
   nghttp2_pq_remove(&src->obq, &stream->pq_entry);
   stream->queued = 0;
@@ -207,33 +232,77 @@ void nghttp2_stream_reschedule(nghttp2_stream *stream) {
   dep_stream = stream->dep_prev;
 
   for (; dep_stream; stream = dep_stream, dep_stream = dep_stream->dep_prev) {
-    if (nghttp2_pq_size(&dep_stream->obq) == 1) {
-      dep_stream->descendant_last_cycle = 0;
-      stream->cycle = 0;
-    } else {
-      /* We update descendant_last_cycle here, and we don't do it when
-         no data is written for stream.  This effectively means that
-         we treat these streams as if they are not scheduled at all.
-         This does not cause disruption in scheduling machinery.  It
-         just makes new streams scheduled a bit early. */
-      dep_stream->descendant_last_cycle = stream->cycle;
+    nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
 
-      nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
+    stream_next_cycle(stream, dep_stream->descendant_last_cycle);
+    stream->seq = dep_stream->descendant_next_seq++;
 
-      stream->cycle =
-          stream_next_cycle(stream, dep_stream->descendant_last_cycle);
-      stream->seq = dep_stream->descendant_next_seq++;
+    nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
 
-      nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
-    }
-
-    DEBUGF(fprintf(stderr, "stream: stream=%d obq resched cycle=%ld\n",
-                   stream->stream_id, stream->cycle));
+    DEBUGF("stream: stream=%d obq resched cycle=%d\n", stream->stream_id,
+           stream->cycle);
 
     dep_stream->last_writelen = stream->last_writelen;
   }
 }
 
+void nghttp2_stream_change_weight(nghttp2_stream *stream, int32_t weight) {
+  nghttp2_stream *dep_stream;
+  uint32_t last_cycle;
+  int32_t old_weight;
+  uint32_t wlen_penalty;
+
+  if (stream->weight == weight) {
+    return;
+  }
+
+  old_weight = stream->weight;
+  stream->weight = weight;
+
+  dep_stream = stream->dep_prev;
+
+  if (!dep_stream) {
+    return;
+  }
+
+  dep_stream->sum_dep_weight += weight - old_weight;
+
+  if (!stream->queued) {
+    return;
+  }
+
+  nghttp2_pq_remove(&dep_stream->obq, &stream->pq_entry);
+
+  wlen_penalty = (uint32_t)stream->last_writelen * NGHTTP2_MAX_WEIGHT;
+
+  /* Compute old stream->pending_penalty we used to calculate
+     stream->cycle */
+  stream->pending_penalty =
+      (uint32_t)((stream->pending_penalty + (uint32_t)old_weight -
+                  (wlen_penalty % (uint32_t)old_weight)) %
+                 (uint32_t)old_weight);
+
+  last_cycle = stream->cycle -
+               (wlen_penalty + stream->pending_penalty) / (uint32_t)old_weight;
+
+  /* Now we have old stream->pending_penalty and new stream->weight in
+     place */
+  stream_next_cycle(stream, last_cycle);
+
+  if (stream->cycle < dep_stream->descendant_last_cycle &&
+      (dep_stream->descendant_last_cycle - stream->cycle) <=
+          NGHTTP2_MAX_CYCLE_DISTANCE) {
+    stream->cycle = dep_stream->descendant_last_cycle;
+  }
+
+  /* Continue to use same stream->seq */
+
+  nghttp2_pq_push(&dep_stream->obq, &stream->pq_entry);
+
+  DEBUGF("stream: stream=%d obq resched cycle=%d\n", stream->stream_id,
+         stream->cycle);
+}
+
 static nghttp2_stream *stream_last_sib(nghttp2_stream *stream) {
   for (; stream->sib_next; stream = stream->sib_next)
     ;
@@ -413,8 +482,7 @@ int nghttp2_stream_attach_item(nghttp2_stream *stream,
   assert((stream->flags & NGHTTP2_STREAM_FLAG_DEFERRED_ALL) == 0);
   assert(stream->item == NULL);
 
-  DEBUGF(fprintf(stderr, "stream: stream=%d attach item=%p\n",
-                 stream->stream_id, item));
+  DEBUGF("stream: stream=%d attach item=%p\n", stream->stream_id, item);
 
   stream->item = item;
 
@@ -432,8 +500,7 @@ int nghttp2_stream_attach_item(nghttp2_stream *stream,
 }
 
 int nghttp2_stream_detach_item(nghttp2_stream *stream) {
-  DEBUGF(fprintf(stderr, "stream: stream=%d detach item=%p\n",
-                 stream->stream_id, stream->item));
+  DEBUGF("stream: stream=%d detach item=%p\n", stream->stream_id, stream->item);
 
   stream->item = NULL;
   stream->flags = (uint8_t)(stream->flags & ~NGHTTP2_STREAM_FLAG_DEFERRED_ALL);
@@ -444,8 +511,8 @@ int nghttp2_stream_detach_item(nghttp2_stream *stream) {
 int nghttp2_stream_defer_item(nghttp2_stream *stream, uint8_t flags) {
   assert(stream->item);
 
-  DEBUGF(fprintf(stderr, "stream: stream=%d defer item=%p cause=%02x\n",
-                 stream->stream_id, stream->item, flags));
+  DEBUGF("stream: stream=%d defer item=%p cause=%02x\n", stream->stream_id,
+         stream->item, flags);
 
   stream->flags |= flags;
 
@@ -455,8 +522,8 @@ int nghttp2_stream_defer_item(nghttp2_stream *stream, uint8_t flags) {
 int nghttp2_stream_resume_deferred_item(nghttp2_stream *stream, uint8_t flags) {
   assert(stream->item);
 
-  DEBUGF(fprintf(stderr, "stream: stream=%d resume item=%p flags=%02x\n",
-                 stream->stream_id, stream->item, flags));
+  DEBUGF("stream: stream=%d resume item=%p flags=%02x\n", stream->stream_id,
+         stream->item, flags);
 
   stream->flags = (uint8_t)(stream->flags & ~flags);
 
@@ -525,9 +592,8 @@ int nghttp2_stream_dep_insert(nghttp2_stream *dep_stream,
   nghttp2_stream *si;
   int rv;
 
-  DEBUGF(fprintf(stderr,
-                 "stream: dep_insert dep_stream(%p)=%d, stream(%p)=%d\n",
-                 dep_stream, dep_stream->stream_id, stream, stream->stream_id));
+  DEBUGF("stream: dep_insert dep_stream(%p)=%d, stream(%p)=%d\n", dep_stream,
+         dep_stream->stream_id, stream, stream->stream_id);
 
   stream->sum_dep_weight = dep_stream->sum_dep_weight;
   dep_stream->sum_dep_weight = stream->weight;
@@ -672,8 +738,8 @@ static void unlink_dep(nghttp2_stream *stream) {
 
 void nghttp2_stream_dep_add(nghttp2_stream *dep_stream,
                             nghttp2_stream *stream) {
-  DEBUGF(fprintf(stderr, "stream: dep_add dep_stream(%p)=%d, stream(%p)=%d\n",
-                 dep_stream, dep_stream->stream_id, stream, stream->stream_id));
+  DEBUGF("stream: dep_add dep_stream(%p)=%d, stream(%p)=%d\n", dep_stream,
+         dep_stream->stream_id, stream, stream->stream_id);
 
   dep_stream->sum_dep_weight += stream->weight;
 
@@ -691,8 +757,7 @@ int nghttp2_stream_dep_remove(nghttp2_stream *stream) {
   int32_t sum_dep_weight_delta;
   int rv;
 
-  DEBUGF(fprintf(stderr, "stream: dep_remove stream(%p)=%d\n", stream,
-                 stream->stream_id));
+  DEBUGF("stream: dep_remove stream(%p)=%d\n", stream, stream->stream_id);
 
   /* Distribute weight of |stream| to direct descendants */
   sum_dep_weight_delta = -stream->weight;
@@ -745,9 +810,8 @@ int nghttp2_stream_dep_insert_subtree(nghttp2_stream *dep_stream,
   nghttp2_stream *si;
   int rv;
 
-  DEBUGF(fprintf(stderr, "stream: dep_insert_subtree dep_stream(%p)=%d "
-                         "stream(%p)=%d\n",
-                 dep_stream, dep_stream->stream_id, stream, stream->stream_id));
+  DEBUGF("stream: dep_insert_subtree dep_stream(%p)=%d stream(%p)=%d\n",
+         dep_stream, dep_stream->stream_id, stream, stream->stream_id);
 
   stream->sum_dep_weight += dep_stream->sum_dep_weight;
   dep_stream->sum_dep_weight = stream->weight;
@@ -794,9 +858,8 @@ int nghttp2_stream_dep_add_subtree(nghttp2_stream *dep_stream,
                                    nghttp2_stream *stream) {
   int rv;
 
-  DEBUGF(fprintf(stderr, "stream: dep_add_subtree dep_stream(%p)=%d "
-                         "stream(%p)=%d\n",
-                 dep_stream, dep_stream->stream_id, stream, stream->stream_id));
+  DEBUGF("stream: dep_add_subtree dep_stream(%p)=%d stream(%p)=%d\n",
+         dep_stream, dep_stream->stream_id, stream, stream->stream_id);
 
   dep_stream->sum_dep_weight += stream->weight;
 
@@ -821,8 +884,8 @@ int nghttp2_stream_dep_add_subtree(nghttp2_stream *dep_stream,
 void nghttp2_stream_dep_remove_subtree(nghttp2_stream *stream) {
   nghttp2_stream *next, *dep_prev;
 
-  DEBUGF(fprintf(stderr, "stream: dep_remove_subtree stream(%p)=%d\n", stream,
-                 stream->stream_id));
+  DEBUGF("stream: dep_remove_subtree stream(%p)=%d\n", stream,
+         stream->stream_id);
 
   assert(stream->dep_prev);
 
@@ -861,9 +924,15 @@ int nghttp2_stream_in_dep_tree(nghttp2_stream *stream) {
 nghttp2_outbound_item *
 nghttp2_stream_next_outbound_item(nghttp2_stream *stream) {
   nghttp2_pq_entry *ent;
+  nghttp2_stream *si;
 
   for (;;) {
     if (stream_active(stream)) {
+      /* Update ascendant's descendant_last_cycle here, so that we can
+         assure that new stream is scheduled based on it. */
+      for (si = stream; si->dep_prev; si = si->dep_prev) {
+        si->dep_prev->descendant_last_cycle = si->cycle;
+      }
       return stream->item;
     }
     ent = nghttp2_pq_top(&stream->obq);
