@@ -350,30 +350,38 @@ void Downstream::crumble_request_cookie(std::vector<nghttp2_nv> &nva) {
 }
 
 namespace {
-void add_header(bool &key_prev, size_t &sum, HeaderRefs &headers,
-                const StringRef &name, const StringRef &value, bool no_index,
-                int32_t token) {
-  key_prev = true;
+void add_header(size_t &sum, HeaderRefs &headers, const StringRef &name,
+                const StringRef &value, bool no_index, int32_t token) {
   sum += name.size() + value.size();
   headers.emplace_back(name, value, no_index, token);
 }
 } // namespace
 
+namespace {
+StringRef alloc_header_name(BlockAllocator &balloc, const StringRef &name) {
+  auto iov = make_byte_ref(balloc, name.size() + 1);
+  auto p = iov.base;
+  p = std::copy(std::begin(name), std::end(name), p);
+  util::inp_strlower(iov.base, p);
+  *p = '\0';
+
+  return StringRef{iov.base, p};
+}
+} // namespace
+
 namespace {
 void append_last_header_key(BlockAllocator &balloc, bool &key_prev, size_t &sum,
                             HeaderRefs &headers, const char *data, size_t len) {
   assert(key_prev);
   sum += len;
   auto &item = headers.back();
-  auto iov = make_byte_ref(balloc, item.name.size() + len + 1);
-  auto p = iov.base;
-  p = std::copy(std::begin(item.name), std::end(item.name), p);
-  p = std::copy_n(data, len, p);
-  util::inp_strlower(p - len, p);
-  *p = '\0';
+  auto name =
+      realloc_concat_string_ref(balloc, item.name, StringRef{data, len});
 
-  item.name = StringRef{iov.base, p};
+  auto p = const_cast<uint8_t *>(name.byte());
+  util::inp_strlower(p + name.size() - len, p + name.size());
 
+  item.name = name;
   item.token = http2::lookup_token(item.name);
 }
 } // namespace
@@ -385,7 +393,8 @@ void append_last_header_value(BlockAllocator &balloc, bool &key_prev,
   key_prev = false;
   sum += len;
   auto &item = headers.back();
-  item.value = concat_string_ref(balloc, item.value, StringRef{data, len});
+  item.value =
+      realloc_concat_string_ref(balloc, item.value, StringRef{data, len});
 }
 } // namespace
 
@@ -435,8 +444,14 @@ const HeaderRefs::value_type *FieldStore::header(const StringRef &name) const {
 
 void FieldStore::add_header_token(const StringRef &name, const StringRef &value,
                                   bool no_index, int32_t token) {
-  shrpx::add_header(header_key_prev_, buffer_size_, headers_, name, value,
-                    no_index, token);
+  shrpx::add_header(buffer_size_, headers_, name, value, no_index, token);
+}
+
+void FieldStore::alloc_add_header_name(const StringRef &name) {
+  auto name_ref = alloc_header_name(balloc_, name);
+  auto token = http2::lookup_token(name_ref);
+  add_header_token(name_ref, StringRef{}, false, token);
+  header_key_prev_ = true;
 }
 
 void FieldStore::append_last_header_key(const char *data, size_t len) {
@@ -449,15 +464,24 @@ void FieldStore::append_last_header_value(const char *data, size_t len) {
                                   headers_, data, len);
 }
 
-void FieldStore::clear_headers() { headers_.clear(); }
+void FieldStore::clear_headers() {
+  headers_.clear();
+  header_key_prev_ = false;
+}
 
 void FieldStore::add_trailer_token(const StringRef &name,
                                    const StringRef &value, bool no_index,
                                    int32_t token) {
   // Header size limit should be applied to all header and trailer
   // fields combined.
-  shrpx::add_header(trailer_key_prev_, buffer_size_, trailers_, name, value,
-                    no_index, token);
+  shrpx::add_header(buffer_size_, trailers_, name, value, no_index, token);
+}
+
+void FieldStore::alloc_add_trailer_name(const StringRef &name) {
+  auto name_ref = alloc_header_name(balloc_, name);
+  auto token = http2::lookup_token(name_ref);
+  add_trailer_token(name_ref, StringRef{}, false, token);
+  trailer_key_prev_ = true;
 }
 
 void FieldStore::append_last_trailer_key(const char *data, size_t len) {
@@ -898,6 +922,10 @@ bool Downstream::get_request_pending() const { return request_pending_; }
 
 void Downstream::set_request_header_sent(bool f) { request_header_sent_ = f; }
 
+bool Downstream::get_request_header_sent() const {
+  return request_header_sent_;
+}
+
 bool Downstream::request_submission_ready() const {
   return (request_state_ == Downstream::HEADER_COMPLETE ||
           request_state_ == Downstream::MSG_COMPLETE) &&
