@@ -108,15 +108,16 @@ int on_stream_close_callback(nghttp2_session *session, int32_t stream_id,
 int Http2Upstream::upgrade_upstream(HttpsUpstream *http) {
   int rv;
 
-  auto http2_settings = http->get_downstream()->get_http2_settings().str();
-  util::to_base64(http2_settings);
+  auto &balloc = http->get_downstream()->get_block_allocator();
 
-  auto settings_payload =
-      base64::decode(std::begin(http2_settings), std::end(http2_settings));
+  auto http2_settings = http->get_downstream()->get_http2_settings();
+  http2_settings = util::to_base64(balloc, http2_settings);
+
+  auto settings_payload = base64::decode(balloc, std::begin(http2_settings),
+                                         std::end(http2_settings));
 
   rv = nghttp2_session_upgrade2(
-      session_, reinterpret_cast<const uint8_t *>(settings_payload.c_str()),
-      settings_payload.size(),
+      session_, settings_payload.byte(), settings_payload.size(),
       http->get_downstream()->request().method == HTTP_HEAD, nullptr);
   if (rv != 0) {
     if (LOG_ENABLED(INFO)) {
@@ -162,8 +163,9 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
                         uint8_t flags, void *user_data) {
   auto namebuf = nghttp2_rcbuf_get_buf(name);
   auto valuebuf = nghttp2_rcbuf_get_buf(value);
+  auto config = get_config();
 
-  if (get_config()->http2.upstream.debug.frame_debug) {
+  if (config->http2.upstream.debug.frame_debug) {
     verbose_on_header_callback(session, frame, namebuf.base, namebuf.len,
                                valuebuf.base, valuebuf.len, flags, user_data);
   }
@@ -179,7 +181,7 @@ int on_header_callback2(nghttp2_session *session, const nghttp2_frame *frame,
 
   auto &req = downstream->request();
 
-  auto &httpconf = get_config()->http;
+  auto &httpconf = config->http;
 
   if (req.fs.buffer_size() + namebuf.len + valuebuf.len >
           httpconf.request_header_field_buffer ||
@@ -277,6 +279,8 @@ int on_begin_headers_callback(nghttp2_session *session,
 
   downstream->reset_upstream_rtimer();
 
+  handler->repeat_read_timer();
+
   auto &req = downstream->request();
 
   // Although, we deprecated minor version from HTTP/2, we supply
@@ -305,10 +309,12 @@ int Http2Upstream::on_request_headers(Downstream *downstream,
       ss << TTY_HTTP_HD << nv.name << TTY_RST << ": " << nv.value << "\n";
     }
     ULOG(INFO, this) << "HTTP request headers. stream_id="
-                     << downstream->get_stream_id() << "\n" << ss.str();
+                     << downstream->get_stream_id() << "\n"
+                     << ss.str();
   }
 
-  auto &dump = get_config()->http2.upstream.debug.dump;
+  auto config = get_config();
+  auto &dump = config->http2.upstream.debug.dump;
 
   if (dump.request_header) {
     http2::dump_nv(dump.request_header, nva);
@@ -337,8 +343,8 @@ int Http2Upstream::on_request_headers(Downstream *downstream,
   auto faddr = handler_->get_upstream_addr();
 
   // For HTTP/2 proxy, we require :authority.
-  if (method_token != HTTP_CONNECT && get_config()->http2_proxy &&
-      !faddr->alt_mode && !authority) {
+  if (method_token != HTTP_CONNECT && config->http2_proxy && !faddr->alt_mode &&
+      !authority) {
     rst_stream(downstream, NGHTTP2_PROTOCOL_ERROR);
     return 0;
   }
@@ -362,7 +368,7 @@ int Http2Upstream::on_request_headers(Downstream *downstream,
     if (method_token == HTTP_OPTIONS &&
         path->value == StringRef::from_lit("*")) {
       // Server-wide OPTIONS request.  Path is empty.
-    } else if (get_config()->http2_proxy && !faddr->alt_mode) {
+    } else if (config->http2_proxy && !faddr->alt_mode) {
       req.path = path->value;
     } else {
       req.path = http2::rewrite_clean_path(downstream->get_block_allocator(),
@@ -469,6 +475,7 @@ int on_frame_recv_callback(nghttp2_session *session, const nghttp2_frame *frame,
     verbose_on_frame_recv_callback(session, frame, user_data);
   }
   auto upstream = static_cast<Http2Upstream *>(user_data);
+  auto handler = upstream->get_client_handler();
 
   switch (frame->hd.type) {
   case NGHTTP2_DATA: {
@@ -502,6 +509,8 @@ int on_frame_recv_callback(nghttp2_session *session, const nghttp2_frame *frame,
     if (frame->headers.cat == NGHTTP2_HCAT_REQUEST) {
       downstream->reset_upstream_rtimer();
 
+      handler->stop_read_timer();
+
       return upstream->on_request_headers(downstream, frame);
     }
 
@@ -898,12 +907,14 @@ nghttp2_session_callbacks *create_http2_upstream_callbacks() {
   nghttp2_session_callbacks_set_send_data_callback(callbacks,
                                                    send_data_callback);
 
-  if (get_config()->padding) {
+  auto config = get_config();
+
+  if (config->padding) {
     nghttp2_session_callbacks_set_select_padding_callback(
         callbacks, http::select_padding_callback);
   }
 
-  if (get_config()->http2.upstream.debug.frame_debug) {
+  if (config->http2.upstream.debug.frame_debug) {
     nghttp2_session_callbacks_set_error_callback(callbacks,
                                                  verbose_error_callback);
   }
@@ -932,7 +943,8 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
       max_buffer_size_(MAX_BUFFER_SIZE) {
   int rv;
 
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &http2conf = config->http2;
 
   auto faddr = handler_->get_upstream_addr();
 
@@ -1025,7 +1037,7 @@ Http2Upstream::Http2Upstream(ClientHandler *handler)
 #endif // defined(TCP_INFO) && defined(TCP_NOTSENT_LOWAT)
 
   handler_->reset_upstream_read_timeout(
-      get_config()->conn.upstream.timeout.http2_read);
+      config->conn.upstream.timeout.http2_read);
 
   handler_->signal_write();
 }
@@ -1074,7 +1086,8 @@ int Http2Upstream::on_read() {
 // After this function call, downstream may be deleted.
 int Http2Upstream::on_write() {
   int rv;
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &http2conf = config->http2;
 
   if ((http2conf.upstream.optimize_write_buffer_size ||
        http2conf.upstream.optimize_window_size) &&
@@ -1199,7 +1212,7 @@ int Http2Upstream::downstream_write(DownstreamConnection *dconn) {
     return downstream_error(dconn, Downstream::EVENT_ERROR);
   }
   if (rv != 0) {
-    return -1;
+    return rv;
   }
   return 0;
 }
@@ -1399,7 +1412,8 @@ int Http2Upstream::send_reply(Downstream *downstream, const uint8_t *body,
   }
 
   const auto &resp = downstream->response();
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
 
   auto &balloc = downstream->get_block_allocator();
 
@@ -1429,8 +1443,7 @@ int Http2Upstream::send_reply(Downstream *downstream, const uint8_t *body,
   }
 
   if (!resp.fs.header(http2::HD_SERVER)) {
-    nva.push_back(http2::make_nv_ls_nocopy(
-        "server", StringRef{get_config()->http.server_name}));
+    nva.push_back(http2::make_nv_ls_nocopy("server", config->http.server_name));
   }
 
   for (auto &p : httpconf.add_response_headers) {
@@ -1476,13 +1489,12 @@ int Http2Upstream::error_reply(Downstream *downstream,
 
   auto response_status = http2::stringify_status(balloc, status_code);
   auto content_length = util::make_string_ref_uint(balloc, html.size());
-  auto date = make_string_ref(balloc, StringRef{lgconf->time_http_str});
+  auto date = make_string_ref(balloc, lgconf->time_http);
 
   auto nva = std::array<nghttp2_nv, 5>{
       {http2::make_nv_ls_nocopy(":status", response_status),
        http2::make_nv_ll("content-type", "text/html; charset=UTF-8"),
-       http2::make_nv_ls_nocopy("server",
-                                StringRef{get_config()->http.server_name}),
+       http2::make_nv_ls_nocopy("server", get_config()->http.server_name),
        http2::make_nv_ls_nocopy("content-length", content_length),
        http2::make_nv_ls_nocopy("date", date)}};
 
@@ -1500,8 +1512,6 @@ int Http2Upstream::error_reply(Downstream *downstream,
 void Http2Upstream::add_pending_downstream(
     std::unique_ptr<Downstream> downstream) {
   downstream_queue_.add_pending(std::move(downstream));
-
-  handler_->stop_read_timer();
 }
 
 void Http2Upstream::remove_downstream(Downstream *downstream) {
@@ -1542,9 +1552,10 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
     }
   }
 
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
 
-  if (!get_config()->http2_proxy && !httpconf.no_location_rewrite) {
+  if (!config->http2_proxy && !httpconf.no_location_rewrite) {
     downstream->rewrite_location_response_header(req.scheme);
   }
 
@@ -1567,7 +1578,7 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
   }
 #endif // HAVE_MRUBY
 
-  auto &http2conf = get_config()->http2;
+  auto &http2conf = config->http2;
 
   // We need some conditions that must be fulfilled to initiate server
   // push.
@@ -1585,7 +1596,7 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
   if (!http2conf.no_server_push &&
       nghttp2_session_get_remote_settings(session_,
                                           NGHTTP2_SETTINGS_ENABLE_PUSH) == 1 &&
-      !get_config()->http2_proxy && (downstream->get_stream_id() % 2) &&
+      !config->http2_proxy && (downstream->get_stream_id() % 2) &&
       resp.fs.header(http2::HD_LINK) &&
       (downstream->get_non_final_response() || resp.http_status == 200) &&
       (req.method == HTTP_GET || req.method == HTTP_POST)) {
@@ -1606,7 +1617,7 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
   nva.push_back(http2::make_nv_ls_nocopy(":status", response_status));
 
   if (downstream->get_non_final_response()) {
-    http2::copy_headers_to_nva(nva, resp.fs.headers());
+    http2::copy_headers_to_nva_nocopy(nva, resp.fs.headers());
 
     if (LOG_ENABLED(INFO)) {
       log_response_headers(downstream, nva);
@@ -1628,9 +1639,8 @@ int Http2Upstream::on_downstream_header_complete(Downstream *downstream) {
 
   http2::copy_headers_to_nva_nocopy(nva, resp.fs.headers());
 
-  if (!get_config()->http2_proxy && !httpconf.no_server_rewrite) {
-    nva.push_back(
-        http2::make_nv_ls_nocopy("server", StringRef{httpconf.server_name}));
+  if (!config->http2_proxy && !httpconf.no_server_rewrite) {
+    nva.push_back(http2::make_nv_ls_nocopy("server", httpconf.server_name));
   } else {
     auto server = resp.fs.header(http2::HD_SERVER);
     if (server) {
@@ -1805,7 +1815,8 @@ void Http2Upstream::log_response_headers(
        << StringRef{nv.value, nv.valuelen} << "\n";
   }
   ULOG(INFO, this) << "HTTP response headers. stream_id="
-                   << downstream->get_stream_id() << "\n" << ss.str();
+                   << downstream->get_stream_id() << "\n"
+                   << ss.str();
 }
 
 int Http2Upstream::on_timeout(Downstream *downstream) {
@@ -1843,6 +1854,11 @@ int Http2Upstream::on_downstream_reset(Downstream *downstream, bool no_retry) {
   }
 
   if (!downstream->request_submission_ready()) {
+    if (downstream->get_response_state() == Downstream::MSG_COMPLETE) {
+      // We have got all response body already.  Send it off.
+      downstream->pop_downstream_connection();
+      return 0;
+    }
     // pushed stream is handled here
     rst_stream(downstream, NGHTTP2_INTERNAL_ERROR);
     downstream->pop_downstream_connection();
@@ -1992,17 +2008,19 @@ int Http2Upstream::submit_push_promise(const StringRef &scheme,
          << StringRef{nv.value, nv.valuelen} << "\n";
     }
     ULOG(INFO, this) << "HTTP push request headers. promised_stream_id="
-                     << promised_stream_id << "\n" << ss.str();
+                     << promised_stream_id << "\n"
+                     << ss.str();
   }
 
   return 0;
 }
 
 bool Http2Upstream::push_enabled() const {
-  return !(get_config()->http2.no_server_push ||
+  auto config = get_config();
+  return !(config->http2.no_server_push ||
            nghttp2_session_get_remote_settings(
                session_, NGHTTP2_SETTINGS_ENABLE_PUSH) == 0 ||
-           get_config()->http2_proxy);
+           config->http2_proxy);
 }
 
 int Http2Upstream::initiate_push(Downstream *downstream, const StringRef &uri) {
