@@ -48,6 +48,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Time out";
   }
@@ -71,11 +75,12 @@ void connect_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto addr = dconn->get_addr();
+  auto raddr = dconn->get_raddr();
 
   DCLOG(WARN, dconn) << "Connect time out; addr="
-                     << util::to_numeric_addr(&addr->addr);
+                     << util::to_numeric_addr(raddr);
 
-  downstream_failure(addr);
+  downstream_failure(addr, raddr);
 
   auto downstream = dconn->get_downstream();
   auto upstream = downstream->get_upstream();
@@ -112,15 +117,51 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
 }
 } // namespace
 
+namespace {
+void backend_retry(Downstream *downstream) {
+  auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
+
+  downstream->add_retry();
+
+  if (downstream->no_more_retry()) {
+    delete handler;
+    return;
+  }
+
+  downstream->pop_downstream_connection();
+
+  auto ndconn = handler->get_downstream_connection(downstream);
+  if (ndconn) {
+    if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
+      return;
+    }
+  }
+
+  downstream->set_request_state(Downstream::CONNECT_FAIL);
+
+  if (upstream->on_downstream_abort_request(downstream, 503) != 0) {
+    delete handler;
+  }
+}
+} // namespace
+
 namespace {
 void writecb(struct ev_loop *loop, ev_io *w, int revents) {
+  int rv;
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
   auto upstream = downstream->get_upstream();
   auto handler = upstream->get_client_handler();
 
-  if (upstream->downstream_write(dconn) != 0) {
+  rv = upstream->downstream_write(dconn);
+  if (rv == SHRPX_ERR_RETRY) {
+    backend_retry(downstream);
+    return;
+  }
+
+  if (rv != 0) {
     delete handler;
   }
 }
@@ -131,23 +172,8 @@ void connectcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
-  auto upstream = downstream->get_upstream();
-  auto handler = upstream->get_client_handler();
   if (dconn->connected() != 0) {
-    downstream->pop_downstream_connection();
-
-    auto ndconn = handler->get_downstream_connection(downstream);
-    if (ndconn) {
-      if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
-        return;
-      }
-    }
-
-    downstream->set_request_state(Downstream::CONNECT_FAIL);
-
-    if (upstream->on_downstream_abort_request(downstream, 503) != 0) {
-      delete handler;
-    }
+    backend_retry(downstream);
     return;
   }
   writecb(loop, w, revents);
@@ -170,21 +196,45 @@ HttpDownstreamConnection::HttpDownstreamConnection(
       ssl_ctx_(worker->get_cl_ssl_ctx()),
       group_(group),
       addr_(nullptr),
+      raddr_(nullptr),
       ioctrl_(&conn_.rlimit),
       response_htp_{0},
-      initial_addr_idx_(initial_addr_idx) {}
+      initial_addr_idx_(initial_addr_idx),
+      reuse_first_write_done_(true),
+      reusable_(true) {}
 
 HttpDownstreamConnection::~HttpDownstreamConnection() {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Deleted";
   }
+
+  if (dns_query_) {
+    auto dns_tracker = worker_->get_dns_tracker();
+    dns_tracker->cancel(dns_query_.get());
+  }
 }
 
 int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
+  int rv;
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
 
+  downstream_ = downstream;
+
+  rv = initiate_connection();
+  if (rv != 0) {
+    downstream_ = nullptr;
+    return rv;
+  }
+
+  return 0;
+}
+
+int HttpDownstreamConnection::initiate_connection() {
+  int rv;
+
   auto worker_blocker = worker_->get_connect_blocker();
   if (worker_blocker->blocked()) {
     if (LOG_ENABLED(INFO)) {
@@ -208,42 +258,117 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
         shared_addr->affinity == AFFINITY_NONE ? shared_addr->next : temp_idx;
     auto end = next_downstream;
     for (;;) {
-      auto &addr = addrs[next_downstream];
+      auto check_dns_result = dns_query_.get() != nullptr;
 
-      if (++next_downstream >= addrs.size()) {
-        next_downstream = 0;
-      }
+      DownstreamAddr *addr;
+      if (check_dns_result) {
+        addr = addr_;
+        addr_ = nullptr;
+        assert(addr);
+        assert(addr->dns);
+      } else {
+        assert(addr_ == nullptr);
+        addr = &addrs[next_downstream];
 
-      if (addr.proto != PROTO_HTTP1) {
-        if (end == next_downstream) {
-          return SHRPX_ERR_NETWORK;
+        if (++next_downstream >= addrs.size()) {
+          next_downstream = 0;
         }
 
-        continue;
+        if (addr->proto != PROTO_HTTP1) {
+          if (end == next_downstream) {
+            return SHRPX_ERR_NETWORK;
+          }
+
+          continue;
+        }
       }
 
-      auto &connect_blocker = addr.connect_blocker;
+      auto &connect_blocker = addr->connect_blocker;
 
       if (connect_blocker->blocked()) {
         if (LOG_ENABLED(INFO)) {
-          DCLOG(INFO, this) << "Backend server "
-                            << util::to_numeric_addr(&addr.addr)
-                            << " was not available temporarily";
+          DCLOG(INFO, this) << "Backend server " << addr->host << ":"
+                            << addr->port << " was not available temporarily";
         }
 
-        if (end == next_downstream) {
+        if (check_dns_result) {
+          dns_query_.reset();
+        } else if (end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
         continue;
       }
 
-      conn_.fd = util::create_nonblock_socket(addr.addr.su.storage.ss_family);
+      Address *raddr;
+
+      if (addr->dns) {
+        if (!check_dns_result) {
+          auto dns_query = make_unique<DNSQuery>(
+              addr->host, [this](int status, const Address *result) {
+                int rv;
+
+                if (status == DNS_STATUS_OK) {
+                  *this->resolved_addr_ = *result;
+                }
+
+                rv = this->initiate_connection();
+                if (rv != 0) {
+                  // This callback destroys |this|.
+                  auto downstream = this->downstream_;
+                  backend_retry(downstream);
+                }
+              });
+
+          auto dns_tracker = worker_->get_dns_tracker();
+
+          if (!resolved_addr_) {
+            resolved_addr_ = make_unique<Address>();
+          }
+          rv = dns_tracker->resolve(resolved_addr_.get(), dns_query.get());
+          switch (rv) {
+          case DNS_STATUS_ERROR:
+            downstream_failure(addr, nullptr);
+            if (end == next_downstream) {
+              return SHRPX_ERR_NETWORK;
+            }
+            continue;
+          case DNS_STATUS_RUNNING:
+            dns_query_ = std::move(dns_query);
+            // Remember current addr
+            addr_ = addr;
+            return 0;
+          case DNS_STATUS_OK:
+            break;
+          default:
+            assert(0);
+          }
+        } else {
+          switch (dns_query_->status) {
+          case DNS_STATUS_ERROR:
+            dns_query_.reset();
+            downstream_failure(addr, nullptr);
+            continue;
+          case DNS_STATUS_OK:
+            dns_query_.reset();
+            break;
+          default:
+            assert(0);
+          }
+        }
+
+        raddr = resolved_addr_.get();
+        util::set_port(*resolved_addr_, addr->port);
+      } else {
+        raddr = &addr->addr;
+      }
+
+      conn_.fd = util::create_nonblock_socket(raddr->su.storage.ss_family);
 
       if (conn_.fd == -1) {
         auto error = errno;
         DCLOG(WARN, this) << "socket() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
         worker_blocker->on_failure();
@@ -253,20 +378,19 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 
       worker_blocker->on_success();
 
-      int rv;
-      rv = connect(conn_.fd, &addr.addr.su.sa, addr.addr.len);
+      rv = connect(conn_.fd, &raddr->su.sa, raddr->len);
       if (rv != 0 && errno != EINPROGRESS) {
         auto error = errno;
         DCLOG(WARN, this) << "connect() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
-        downstream_failure(&addr);
+        downstream_failure(addr, raddr);
 
         close(conn_.fd);
         conn_.fd = -1;
 
-        if (end == next_downstream) {
+        if (!check_dns_result && end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
@@ -278,7 +402,8 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
         DCLOG(INFO, this) << "Connecting to downstream server";
       }
 
-      addr_ = &addr;
+      addr_ = addr;
+      raddr_ = raddr;
 
       if (addr_->tls) {
         assert(ssl_ctx_);
@@ -315,18 +440,23 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
       break;
     }
 
-    // TODO we should have timeout for connection establishment
+    conn_.wt.repeat = downstreamconf.timeout.connect;
     ev_timer_again(conn_.loop, &conn_.wt);
   } else {
     // we may set read timer cb to idle_timeoutcb.  Reset again.
-    conn_.rt.repeat = downstreamconf.timeout.read;
     ev_set_cb(&conn_.rt, timeoutcb);
-    ev_timer_stop(conn_.loop, &conn_.rt);
+    if (conn_.read_timeout < downstreamconf.timeout.read) {
+      conn_.read_timeout = downstreamconf.timeout.read;
+      conn_.last_read = ev_now(conn_.loop);
+    } else {
+      conn_.again_rt(downstreamconf.timeout.read);
+    }
 
     ev_set_cb(&conn_.rev, readcb);
-  }
 
-  downstream_ = downstream;
+    do_write_ = &HttpDownstreamConnection::write_reuse_first;
+    reuse_first_write_done_ = false;
+  }
 
   http_parser_init(&response_htp_, HTTP_RESPONSE);
   response_htp_.data = downstream_;
@@ -335,6 +465,10 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 }
 
 int HttpDownstreamConnection::push_request_headers() {
+  if (downstream_->get_request_header_sent()) {
+    return 0;
+  }
+
   const auto &downstream_hostport = addr_->hostport;
   const auto &req = downstream_->request();
 
@@ -342,7 +476,8 @@ int HttpDownstreamConnection::push_request_headers() {
 
   auto connect_method = req.method == HTTP_CONNECT;
 
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
 
   // Set request_sent to true because we write request into buffer
   // here.
@@ -352,7 +487,7 @@ int HttpDownstreamConnection::push_request_headers() {
   // case, we use backend server's host nonetheless.
   auto authority = StringRef(downstream_hostport);
   auto no_host_rewrite =
-      httpconf.no_host_rewrite || get_config()->http2_proxy || connect_method;
+      httpconf.no_host_rewrite || config->http2_proxy || connect_method;
 
   if (no_host_rewrite && !req.authority.empty()) {
     authority = req.authority;
@@ -365,11 +500,11 @@ int HttpDownstreamConnection::push_request_headers() {
   // Assume that method and request path do not contain \r\n.
   auto meth = http2::to_method_string(req.method);
   buf->append(meth);
-  buf->append(" ");
+  buf->append(' ');
 
   if (connect_method) {
     buf->append(authority);
-  } else if (get_config()->http2_proxy) {
+  } else if (config->http2_proxy) {
     // Construct absolute-form request target because we are going to
     // send a request to a HTTP/1 proxy.
     assert(!req.scheme.empty());
@@ -434,7 +569,7 @@ int HttpDownstreamConnection::push_request_headers() {
   if (fwdconf.params) {
     auto params = fwdconf.params;
 
-    if (get_config()->http2_proxy || connect_method) {
+    if (config->http2_proxy || connect_method) {
       params &= ~FORWARDED_PROTO;
     }
 
@@ -478,7 +613,7 @@ int HttpDownstreamConnection::push_request_headers() {
     buf->append((*xff).value);
     buf->append("\r\n");
   }
-  if (!get_config()->http2_proxy && !connect_method) {
+  if (!config->http2_proxy && !connect_method) {
     buf->append("X-Forwarded-Proto: ");
     assert(!req.scheme.empty());
     buf->append(req.scheme);
@@ -522,7 +657,8 @@ int HttpDownstreamConnection::push_request_headers() {
       nhdrs = http::colorizeHeaders(nhdrs.c_str());
     }
     DCLOG(INFO, this) << "HTTP request headers. stream_id="
-                      << downstream_->get_stream_id() << "\n" << nhdrs;
+                      << downstream_->get_stream_id() << "\n"
+                      << nhdrs;
   }
 
   // Don't call signal_write() if we anticipate request body.  We call
@@ -559,6 +695,8 @@ int HttpDownstreamConnection::push_upload_data_chunk(const uint8_t *data,
 }
 
 int HttpDownstreamConnection::end_upload_data() {
+  signal_write();
+
   if (!downstream_->get_chunked_request()) {
     return 0;
   }
@@ -575,8 +713,6 @@ int HttpDownstreamConnection::end_upload_data() {
     output->append("\r\n");
   }
 
-  signal_write();
-
   return 0;
 }
 
@@ -615,6 +751,9 @@ namespace {
 void idle_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
+
+  // We don't have to check conn->expired_rt() since we restart timer
+  // when connection gets idle.
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Idle connection timeout";
   }
@@ -635,9 +774,13 @@ void HttpDownstreamConnection::detach_downstream(Downstream *downstream) {
 
   auto &downstreamconf = *worker_->get_downstream_config();
 
-  conn_.rt.repeat = downstreamconf.timeout.idle_read;
   ev_set_cb(&conn_.rt, idle_timeoutcb);
-  ev_timer_again(conn_.loop, &conn_.rt);
+  if (conn_.read_timeout < downstreamconf.timeout.idle_read) {
+    conn_.read_timeout = downstreamconf.timeout.idle_read;
+    conn_.last_read = ev_now(conn_.loop);
+  } else {
+    conn_.again_rt(downstreamconf.timeout.idle_read);
+  }
 
   conn_.wlimit.stopw();
   ev_timer_stop(conn_.loop, &conn_.wt);
@@ -699,7 +842,37 @@ int htp_hdrs_completecb(http_parser *htp) {
   downstream->set_downstream_addr_group(dconn->get_downstream_addr_group());
   downstream->set_addr(dconn->get_addr());
 
-  if (resp.fs.parse_content_length() != 0) {
+  // Server MUST NOT send Transfer-Encoding with a status code 1xx or
+  // 204.  Also server MUST NOT send Transfer-Encoding with a status
+  // code 200 to a CONNECT request.  Same holds true with
+  // Content-Length.
+  if (resp.http_status == 204) {
+    if (resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+    // Some server send content-length: 0 for 204.  Until they get
+    // fixed, we accept, but ignore it.
+
+    // Calling parse_content_length() detects duplicated
+    // content-length header fields.
+    if (resp.fs.parse_content_length() != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length == 0) {
+      auto cl = resp.fs.header(http2::HD_CONTENT_LENGTH);
+      assert(cl);
+      http2::erase_header(cl);
+    }
+  } else if (resp.http_status / 100 == 1 ||
+             (resp.http_status == 200 && req.method == HTTP_CONNECT)) {
+    if (resp.fs.header(http2::HD_CONTENT_LENGTH) ||
+        resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+  } else if (resp.fs.parse_content_length() != 0) {
     downstream->set_response_state(Downstream::MSG_BAD_HEADER);
     return -1;
   }
@@ -790,8 +963,8 @@ int ensure_max_header_fields(const Downstream *downstream,
 
   if (resp.fs.num_fields() >= httpconf.max_response_header_fields) {
     if (LOG_ENABLED(INFO)) {
-      DLOG(INFO, downstream)
-          << "Too many header field num=" << resp.fs.num_fields() + 1;
+      DLOG(INFO, downstream) << "Too many header field num="
+                             << resp.fs.num_fields() + 1;
     }
     return -1;
   }
@@ -805,7 +978,6 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
   auto downstream = static_cast<Downstream *>(htp->data);
   auto &resp = downstream->response();
   auto &httpconf = get_config()->http;
-  auto &balloc = downstream->get_block_allocator();
 
   if (ensure_header_field_buffer(downstream, httpconf, len) != 0) {
     return -1;
@@ -818,9 +990,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
       if (ensure_max_header_fields(downstream, httpconf) != 0) {
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_header_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_header_name(StringRef{data, len});
     }
   } else {
     // trailer part
@@ -833,9 +1003,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
         // wrong place or crash if trailer fields are currently empty.
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_trailer_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_trailer_name(StringRef{data, len});
     }
   }
   return 0;
@@ -914,7 +1082,33 @@ http_parser_settings htp_hooks = {
 };
 } // namespace
 
+int HttpDownstreamConnection::write_reuse_first() {
+  int rv;
+
+  if (conn_.tls.ssl) {
+    rv = write_tls();
+  } else {
+    rv = write_clear();
+  }
+
+  if (rv != 0) {
+    return SHRPX_ERR_RETRY;
+  }
+
+  if (conn_.tls.ssl) {
+    do_write_ = &HttpDownstreamConnection::write_tls;
+  } else {
+    do_write_ = &HttpDownstreamConnection::write_clear;
+  }
+
+  reuse_first_write_done_ = true;
+
+  return 0;
+}
+
 int HttpDownstreamConnection::read_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<uint8_t, 16_k> buf;
   int rv;
 
@@ -940,6 +1134,8 @@ int HttpDownstreamConnection::read_clear() {
 }
 
 int HttpDownstreamConnection::write_clear() {
+  conn_.last_read = ev_now(conn_.loop);
+
   auto upstream = downstream_->get_upstream();
   auto input = downstream_->get_request_buf();
 
@@ -955,7 +1151,16 @@ int HttpDownstreamConnection::write_clear() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      do_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -977,7 +1182,7 @@ int HttpDownstreamConnection::write_clear() {
 int HttpDownstreamConnection::tls_handshake() {
   ERR_clear_error();
 
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   auto rv = conn_.tls_handshake();
   if (rv == SHRPX_ERR_INPROGRESS) {
@@ -985,7 +1190,7 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (rv < 0) {
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return rv;
   }
@@ -995,8 +1200,8 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (!get_config()->tls.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_) != 0) {
-    downstream_failure(addr_);
+      ssl::check_cert(conn_.tls.ssl, addr_, raddr_) != 0) {
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1004,7 +1209,7 @@ int HttpDownstreamConnection::tls_handshake() {
   if (!SSL_session_reused(conn_.tls.ssl)) {
     auto session = SSL_get0_session(conn_.tls.ssl);
     if (session) {
-      ssl::try_cache_tls_session(addr_->tls_session_cache, addr_->addr, session,
+      ssl::try_cache_tls_session(addr_->tls_session_cache, *raddr_, session,
                                  ev_now(conn_.loop));
     }
   }
@@ -1027,6 +1232,8 @@ int HttpDownstreamConnection::tls_handshake() {
 }
 
 int HttpDownstreamConnection::read_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
   std::array<uint8_t, 16_k> buf;
@@ -1054,6 +1261,8 @@ int HttpDownstreamConnection::read_tls() {
 }
 
 int HttpDownstreamConnection::write_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
   auto upstream = downstream_->get_upstream();
@@ -1071,7 +1280,16 @@ int HttpDownstreamConnection::write_tls() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      do_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -1165,10 +1383,10 @@ int HttpDownstreamConnection::connected() {
     conn_.wlimit.stopw();
 
     DCLOG(WARN, this) << "Backend connect failed; addr="
-                      << util::to_numeric_addr(&addr_->addr)
+                      << util::to_numeric_addr(raddr_)
                       << ": errno=" << sock_error;
 
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1177,7 +1395,14 @@ int HttpDownstreamConnection::connected() {
     DCLOG(INFO, this) << "Connected to downstream host";
   }
 
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  // Reset timeout for write.  Previously, we set timeout for connect.
+  conn_.wt.repeat = downstreamconf.timeout.write;
+  ev_timer_again(conn_.loop, &conn_.wt);
+
   conn_.rlimit.startw();
+  conn_.again_rt();
 
   ev_set_cb(&conn_.wev, writecb);
 
@@ -1223,6 +1448,10 @@ HttpDownstreamConnection::get_downstream_addr_group() const {
 
 DownstreamAddr *HttpDownstreamConnection::get_addr() const { return addr_; }
 
-bool HttpDownstreamConnection::poolable() const { return !group_->retired; }
+bool HttpDownstreamConnection::poolable() const {
+  return !group_->retired && reusable_;
+}
+
+const Address *HttpDownstreamConnection::get_raddr() const { return raddr_; }
 
 } // namespace shrpx
