@@ -107,45 +107,67 @@ mrb_value response_mod_header(mrb_state *mrb, mrb_value self, bool repl) {
   auto data = static_cast<MRubyAssocData *>(mrb->ud);
   auto downstream = data->downstream;
   auto &resp = downstream->response();
+  auto &balloc = downstream->get_block_allocator();
 
   mrb_value key, values;
-  mrb_get_args(mrb, "oo", &key, &values);
+  mrb_get_args(mrb, "So", &key, &values);
 
   if (RSTRING_LEN(key) == 0) {
     mrb_raise(mrb, E_RUNTIME_ERROR, "empty key is not allowed");
   }
 
+  auto ai = mrb_gc_arena_save(mrb);
+
   key = mrb_funcall(mrb, key, "downcase", 0);
 
+  auto keyref =
+      make_string_ref(balloc, StringRef{RSTRING_PTR(key),
+                                        static_cast<size_t>(RSTRING_LEN(key))});
+
+  mrb_gc_arena_restore(mrb, ai);
+
+  auto token = http2::lookup_token(keyref.byte(), keyref.size());
+
   if (repl) {
     size_t p = 0;
     auto &headers = resp.fs.headers();
     for (size_t i = 0; i < headers.size(); ++i) {
-      auto &hd = headers[i];
-      if (util::streq(std::begin(hd.name), hd.name.size(), RSTRING_PTR(key),
-                      RSTRING_LEN(key))) {
+      auto &kv = headers[i];
+      if (kv.name == keyref) {
         continue;
       }
       if (i != p) {
-        headers[p++] = std::move(hd);
+        headers[p++] = std::move(kv);
       }
     }
     headers.resize(p);
   }
 
-  if (mrb_obj_is_instance_of(mrb, values, mrb->array_class)) {
+  if (mrb_array_p(values)) {
     auto n = mrb_ary_len(mrb, values);
     for (int i = 0; i < n; ++i) {
-      auto value = mrb_ary_entry(values, i);
-      resp.fs.add_header(std::string(RSTRING_PTR(key), RSTRING_LEN(key)),
-                         std::string(RSTRING_PTR(value), RSTRING_LEN(value)));
-    }
-  } else if (!mrb_nil_p(values)) {
-    resp.fs.add_header(std::string(RSTRING_PTR(key), RSTRING_LEN(key)),
-                       std::string(RSTRING_PTR(values), RSTRING_LEN(values)));
-  }
+      auto value = mrb_ary_ref(mrb, values, i);
+      if (!mrb_string_p(value)) {
+        mrb_raise(mrb, E_RUNTIME_ERROR, "value must be string");
+      }
 
-  data->response_headers_dirty = true;
+      resp.fs.add_header_token(
+          keyref,
+          make_string_ref(balloc,
+                          StringRef{RSTRING_PTR(value),
+                                    static_cast<size_t>(RSTRING_LEN(value))}),
+          false, token);
+    }
+  } else if (mrb_string_p(values)) {
+    resp.fs.add_header_token(
+        keyref,
+        make_string_ref(balloc,
+                        StringRef{RSTRING_PTR(values),
+                                  static_cast<size_t>(RSTRING_LEN(values))}),
+        false, token);
+  } else {
+    mrb_raise(mrb, E_RUNTIME_ERROR, "value must be string");
+  }
 
   return mrb_nil_value();
 }
@@ -179,9 +201,12 @@ namespace {
 mrb_value response_return(mrb_state *mrb, mrb_value self) {
   auto data = static_cast<MRubyAssocData *>(mrb->ud);
   auto downstream = data->downstream;
+  auto &req = downstream->request();
   auto &resp = downstream->response();
   int rv;
 
+  auto &balloc = downstream->get_block_allocator();
+
   if (downstream->get_response_state() == Downstream::MSG_COMPLETE) {
     mrb_raise(mrb, E_RUNTIME_ERROR, "response has already been committed");
   }
@@ -197,30 +222,41 @@ mrb_value response_return(mrb_state *mrb, mrb_value self) {
     resp.http_status = 200;
   }
 
-  if (data->response_headers_dirty) {
-    resp.fs.index_headers();
-    data->response_headers_dirty = false;
-  }
-
   if (downstream->expect_response_body() && vallen > 0) {
     body = reinterpret_cast<const uint8_t *>(val);
     bodylen = vallen;
   }
 
   auto cl = resp.fs.header(http2::HD_CONTENT_LENGTH);
-  if (cl) {
-    cl->value = util::utos(bodylen);
+
+  if (resp.http_status == 204 ||
+      (resp.http_status == 200 && req.method == HTTP_CONNECT)) {
+    if (cl) {
+      // Delete content-length here
+      http2::erase_header(cl);
+    }
+
+    resp.fs.content_length = -1;
   } else {
-    resp.fs.add_header("content-length", util::utos(bodylen),
-                       http2::HD_CONTENT_LENGTH);
+    auto content_length = util::make_string_ref_uint(balloc, vallen);
+
+    if (cl) {
+      cl->value = content_length;
+    } else {
+      resp.fs.add_header_token(StringRef::from_lit("content-length"),
+                               content_length, false, http2::HD_CONTENT_LENGTH);
+    }
+
+    resp.fs.content_length = vallen;
   }
-  resp.fs.content_length = bodylen;
 
   auto date = resp.fs.header(http2::HD_DATE);
   if (!date) {
     auto lgconf = log_config();
     lgconf->update_tstamp(std::chrono::system_clock::now());
-    resp.fs.add_header("date", lgconf->time_http_str, http2::HD_DATE);
+    resp.fs.add_header_token(StringRef::from_lit("date"),
+                             make_string_ref(balloc, lgconf->time_http), false,
+                             http2::HD_DATE);
   }
 
   auto upstream = downstream->get_upstream();
