@@ -72,12 +72,11 @@ public:
   virtual int on_downstream_body_complete(Downstream *downstream);
 
   virtual void on_handler_delete();
-  virtual int on_downstream_reset(bool no_retry);
+  virtual int on_downstream_reset(Downstream *downstream, bool no_retry);
 
   virtual int send_reply(Downstream *downstream, const uint8_t *body,
                          size_t bodylen);
-  virtual int initiate_push(Downstream *downstream, const char *uri,
-                            size_t len);
+  virtual int initiate_push(Downstream *downstream, const StringRef &uri);
   virtual int response_riovec(struct iovec *iov, int iovcnt) const;
   virtual void response_drain(size_t n);
   virtual bool response_empty() const;
@@ -97,16 +96,13 @@ public:
   void start_downstream(Downstream *downstream);
   void initiate_downstream(Downstream *downstream);
 
-  using WriteBuffer = Buffer<32_k>;
-
-  WriteBuffer *get_response_buf();
+  DefaultMemchunks *get_response_buf();
 
 private:
-  WriteBuffer wb_;
+  DefaultMemchunks wb_;
   DownstreamQueue downstream_queue_;
   ClientHandler *handler_;
   spdylay_session *session_;
-  int32_t initial_window_size_;
   bool flow_control_;
 };
 
