@@ -454,6 +454,60 @@ void test_nghttp2_frame_pack_window_update(void) {
   nghttp2_frame_window_update_free(&frame);
 }
 
+void test_nghttp2_frame_pack_altsvc(void) {
+  nghttp2_extension frame, oframe;
+  nghttp2_ext_altsvc altsvc, oaltsvc;
+  nghttp2_bufs bufs;
+  int rv;
+  size_t payloadlen;
+  static const uint8_t origin[] = "nghttp2.org";
+  static const uint8_t field_value[] = "h2=\":443\"";
+  nghttp2_buf buf;
+  uint8_t *rawbuf;
+  nghttp2_mem *mem;
+
+  mem = nghttp2_mem_default();
+
+  frame_pack_bufs_init(&bufs);
+
+  frame.payload = &altsvc;
+  oframe.payload = &oaltsvc;
+
+  rawbuf = nghttp2_mem_malloc(mem, 32);
+  nghttp2_buf_wrap_init(&buf, rawbuf, 32);
+
+  buf.last = nghttp2_cpymem(buf.last, origin, sizeof(origin) - 1);
+  buf.last = nghttp2_cpymem(buf.last, field_value, sizeof(field_value) - 1);
+
+  nghttp2_frame_altsvc_init(&frame, 1000000007, buf.pos, sizeof(origin) - 1,
+                            buf.pos + sizeof(origin) - 1,
+                            sizeof(field_value) - 1);
+
+  payloadlen = 2 + sizeof(origin) - 1 + sizeof(field_value) - 1;
+
+  rv = nghttp2_frame_pack_altsvc(&bufs, &frame);
+
+  CU_ASSERT(0 == rv);
+  CU_ASSERT(NGHTTP2_FRAME_HDLEN + payloadlen == nghttp2_bufs_len(&bufs));
+
+  rv = unpack_framebuf((nghttp2_frame *)&oframe, &bufs);
+
+  CU_ASSERT(0 == rv);
+
+  check_frame_header(payloadlen, NGHTTP2_ALTSVC, NGHTTP2_FLAG_NONE, 1000000007,
+                     &oframe.hd);
+
+  CU_ASSERT(sizeof(origin) - 1 == oaltsvc.origin_len);
+  CU_ASSERT(0 == memcmp(origin, oaltsvc.origin, sizeof(origin) - 1));
+  CU_ASSERT(sizeof(field_value) - 1 == oaltsvc.field_value_len);
+  CU_ASSERT(0 ==
+            memcmp(field_value, oaltsvc.field_value, sizeof(field_value) - 1));
+
+  nghttp2_frame_altsvc_free(&oframe, mem);
+  nghttp2_frame_altsvc_free(&frame, mem);
+  nghttp2_bufs_free(&bufs);
+}
+
 void test_nghttp2_nv_array_copy(void) {
   nghttp2_nv *nva;
   ssize_t rv;
