@@ -1001,20 +1001,25 @@ void test_nghttp2_hd_deflate_inflate(void) {
   nghttp2_hd_deflater deflater;
   nghttp2_hd_inflater inflater;
   nghttp2_nv nv1[] = {
-      MAKE_NV(":status", "200 OK"), MAKE_NV("access-control-allow-origin", "*"),
+      MAKE_NV(":status", "200 OK"),
+      MAKE_NV("access-control-allow-origin", "*"),
       MAKE_NV("cache-control", "private, max-age=0, must-revalidate"),
-      MAKE_NV("content-length", "76073"), MAKE_NV("content-type", "text/html"),
+      MAKE_NV("content-length", "76073"),
+      MAKE_NV("content-type", "text/html"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
       MAKE_NV("expires", "Sat, 27 Jul 2013 06:22:12 GMT"),
-      MAKE_NV("server", "Apache"), MAKE_NV("vary", "foobar"),
+      MAKE_NV("server", "Apache"),
+      MAKE_NV("vary", "foobar"),
       MAKE_NV("via", "1.1 alphabravo (squid/3.x.x), 1.1 nghttpx"),
       MAKE_NV("x-cache", "MISS from alphabravo"),
-      MAKE_NV("x-cache-action", "MISS"), MAKE_NV("x-cache-age", "0"),
+      MAKE_NV("x-cache-action", "MISS"),
+      MAKE_NV("x-cache-age", "0"),
       MAKE_NV("x-cache-lookup", "MISS from alphabravo:3128"),
       MAKE_NV("x-lb-nocache", "true"),
   };
   nghttp2_nv nv2[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=56682045"),
       MAKE_NV("content-type", "text/css"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1025,7 +1030,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache", "HIT from alphabravo"),
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128")};
   nghttp2_nv nv3[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=56682072"),
       MAKE_NV("content-type", "text/css"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1037,7 +1043,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv4[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=56682022"),
       MAKE_NV("content-type", "text/css"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1049,7 +1056,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv5[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=4461139"),
       MAKE_NV("content-type", "application/x-javascript"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1061,7 +1069,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv6[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=18645951"),
       MAKE_NV("content-type", "application/x-javascript"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1073,7 +1082,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv7[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=31536000"),
       MAKE_NV("content-type", "application/javascript"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1085,7 +1095,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv8[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=31536000"),
       MAKE_NV("content-type", "application/javascript"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1097,7 +1108,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv9[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=31536000"),
       MAKE_NV("content-type", "application/javascript"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1109,7 +1121,8 @@ void test_nghttp2_hd_deflate_inflate(void) {
       MAKE_NV("x-cache-lookup", "HIT from alphabravo:3128"),
   };
   nghttp2_nv nv10[] = {
-      MAKE_NV(":status", "304 Not Modified"), MAKE_NV("age", "0"),
+      MAKE_NV(":status", "304 Not Modified"),
+      MAKE_NV("age", "0"),
       MAKE_NV("cache-control", "max-age=56682045"),
       MAKE_NV("content-type", "text/css"),
       MAKE_NV("date", "Sat, 27 Jul 2013 06:22:12 GMT"),
@@ -1265,6 +1278,144 @@ void test_nghttp2_hd_public_api(void) {
   nghttp2_hd_deflate_del(deflater);
 }
 
+void test_nghttp2_hd_deflate_hd_vec(void) {
+  nghttp2_hd_deflater *deflater;
+  nghttp2_hd_inflater *inflater;
+  nghttp2_nv nva[] = {
+      MAKE_NV(":method", "PUT"),
+      MAKE_NV(":scheme", "https"),
+      MAKE_NV(":authority", "localhost:3000"),
+      MAKE_NV(":path", "/usr/foo/alpha/bravo"),
+      MAKE_NV("content-type", "image/png"),
+      MAKE_NV("content-length", "1000000007"),
+  };
+  uint8_t buf[4096];
+  ssize_t blocklen;
+  nghttp2_mem *mem;
+  nghttp2_vec vec[256];
+  size_t buflen;
+  nghttp2_bufs bufs;
+  nva_out out;
+  size_t i;
+
+  mem = nghttp2_mem_default();
+
+  nva_out_init(&out);
+
+  nghttp2_hd_deflate_new(&deflater, 4096);
+  nghttp2_hd_inflate_new(&inflater);
+
+  buflen = nghttp2_hd_deflate_bound(deflater, nva, ARRLEN(nva));
+
+  vec[0].base = &buf[0];
+  vec[0].len = buflen / 2;
+  vec[1].base = &buf[buflen / 2];
+  vec[1].len = buflen / 2;
+
+  blocklen = nghttp2_hd_deflate_hd_vec(deflater, vec, 2, nva, ARRLEN(nva));
+
+  CU_ASSERT(blocklen > 0);
+
+  nghttp2_bufs_wrap_init(&bufs, buf, (size_t)blocklen, mem);
+  bufs.head->buf.last += blocklen;
+
+  CU_ASSERT(blocklen == inflate_hd(inflater, &out, &bufs, 0, mem));
+
+  CU_ASSERT(ARRLEN(nva) == out.nvlen);
+  assert_nv_equal(nva, out.nva, ARRLEN(nva), mem);
+
+  nghttp2_bufs_wrap_free(&bufs);
+
+  nghttp2_hd_inflate_del(inflater);
+  nghttp2_hd_deflate_del(deflater);
+  nva_out_reset(&out, mem);
+
+  /* check the case when veclen is 0 */
+  nghttp2_hd_deflate_new(&deflater, 4096);
+  nghttp2_hd_inflate_new(&inflater);
+
+  blocklen = nghttp2_hd_deflate_hd_vec(deflater, NULL, 0, nva, ARRLEN(nva));
+
+  CU_ASSERT(NGHTTP2_ERR_INSUFF_BUFSIZE == blocklen);
+
+  nghttp2_hd_inflate_del(inflater);
+  nghttp2_hd_deflate_del(deflater);
+
+  /* check the case when chunk length is 0 */
+  vec[0].base = NULL;
+  vec[0].len = 0;
+  vec[1].base = NULL;
+  vec[1].len = 0;
+
+  nghttp2_hd_deflate_new(&deflater, 4096);
+  nghttp2_hd_inflate_new(&inflater);
+
+  blocklen = nghttp2_hd_deflate_hd_vec(deflater, vec, 2, nva, ARRLEN(nva));
+
+  CU_ASSERT(NGHTTP2_ERR_INSUFF_BUFSIZE == blocklen);
+
+  nghttp2_hd_inflate_del(inflater);
+  nghttp2_hd_deflate_del(deflater);
+
+  /* check the case where chunk size differs in each chunk */
+  nghttp2_hd_deflate_new(&deflater, 4096);
+  nghttp2_hd_inflate_new(&inflater);
+
+  buflen = nghttp2_hd_deflate_bound(deflater, nva, ARRLEN(nva));
+
+  vec[0].base = &buf[0];
+  vec[0].len = buflen / 2;
+  vec[1].base = &buf[buflen / 2];
+  vec[1].len = (buflen / 2) + 1;
+
+  blocklen = nghttp2_hd_deflate_hd_vec(deflater, vec, 2, nva, ARRLEN(nva));
+
+  CU_ASSERT(blocklen > 0);
+
+  nghttp2_bufs_wrap_init(&bufs, buf, (size_t)blocklen, mem);
+  bufs.head->buf.last += blocklen;
+
+  CU_ASSERT(blocklen == inflate_hd(inflater, &out, &bufs, 0, mem));
+  CU_ASSERT(ARRLEN(nva) == out.nvlen);
+  assert_nv_equal(nva, out.nva, ARRLEN(nva), mem);
+
+  nghttp2_bufs_wrap_free(&bufs);
+
+  nghttp2_hd_inflate_del(inflater);
+  nghttp2_hd_deflate_del(deflater);
+  nva_out_reset(&out, mem);
+
+  /* check the case where chunk size is 1 */
+  nghttp2_hd_deflate_new(&deflater, 4096);
+  nghttp2_hd_inflate_new(&inflater);
+
+  buflen = nghttp2_hd_deflate_bound(deflater, nva, ARRLEN(nva));
+
+  assert(buflen <= ARRLEN(vec));
+
+  for (i = 0; i < buflen; ++i) {
+    vec[i].base = &buf[i];
+    vec[i].len = 1;
+  }
+
+  blocklen = nghttp2_hd_deflate_hd_vec(deflater, vec, buflen, nva, ARRLEN(nva));
+
+  CU_ASSERT(blocklen > 0);
+
+  nghttp2_bufs_wrap_init(&bufs, buf, (size_t)blocklen, mem);
+  bufs.head->buf.last += blocklen;
+
+  CU_ASSERT(blocklen == inflate_hd(inflater, &out, &bufs, 0, mem));
+  CU_ASSERT(ARRLEN(nva) == out.nvlen);
+  assert_nv_equal(nva, out.nva, ARRLEN(nva), mem);
+
+  nghttp2_bufs_wrap_free(&bufs);
+
+  nghttp2_hd_inflate_del(inflater);
+  nghttp2_hd_deflate_del(deflater);
+  nva_out_reset(&out, mem);
+}
+
 static size_t encode_length(uint8_t *buf, uint64_t n, size_t prefix) {
   size_t k = (size_t)((1 << prefix) - 1);
   size_t len = 0;
@@ -1296,7 +1447,7 @@ static size_t encode_length(uint8_t *buf, uint64_t n, size_t prefix) {
 void test_nghttp2_hd_decode_length(void) {
   uint32_t out;
   size_t shift;
-  int final;
+  int fin;
   uint8_t buf[16];
   uint8_t *bufp;
   size_t len;
@@ -1306,39 +1457,52 @@ void test_nghttp2_hd_decode_length(void) {
   memset(buf, 0, sizeof(buf));
   len = encode_length(buf, UINT32_MAX, 7);
 
-  rv = nghttp2_hd_decode_length(&out, &shift, &final, 0, 0, buf, buf + len, 7);
+  rv = nghttp2_hd_decode_length(&out, &shift, &fin, 0, 0, buf, buf + len, 7);
 
   CU_ASSERT((ssize_t)len == rv);
-  CU_ASSERT(0 != final);
+  CU_ASSERT(0 != fin);
   CU_ASSERT(UINT32_MAX == out);
 
   /* Make sure that we can decode integer if we feed 1 byte at a
      time */
   out = 0;
   shift = 0;
-  final = 0;
+  fin = 0;
   bufp = buf;
 
   for (i = 0; i < len; ++i, ++bufp) {
-    rv = nghttp2_hd_decode_length(&out, &shift, &final, out, shift, bufp,
+    rv = nghttp2_hd_decode_length(&out, &shift, &fin, out, shift, bufp,
                                   bufp + 1, 7);
 
     CU_ASSERT(rv == 1);
 
-    if (final) {
+    if (fin) {
       break;
     }
   }
 
   CU_ASSERT(i == len - 1);
-  CU_ASSERT(0 != final);
+  CU_ASSERT(0 != fin);
   CU_ASSERT(UINT32_MAX == out);
 
   /* Check overflow case */
   memset(buf, 0, sizeof(buf));
   len = encode_length(buf, 1ll << 32, 7);
 
-  rv = nghttp2_hd_decode_length(&out, &shift, &final, 0, 0, buf, buf + len, 7);
+  rv = nghttp2_hd_decode_length(&out, &shift, &fin, 0, 0, buf, buf + len, 7);
+
+  CU_ASSERT(-1 == rv);
+
+  /* Check the case that shift goes beyond 32 bits */
+  buf[0] = 255;
+  buf[1] = 128;
+  buf[2] = 128;
+  buf[3] = 128;
+  buf[4] = 128;
+  buf[5] = 128;
+  buf[6] = 1;
+
+  rv = nghttp2_hd_decode_length(&out, &shift, &fin, 0, 0, buf, buf + 7, 8);
 
   CU_ASSERT(-1 == rv);
 }
